#define ZASM_LINK_NAMESPACE_C 1

#include <zasm/link_namespace.h>
#include <zasm/namespace.h>

struct link_namespace_t *link_namespace_new()
{
  return LINK_NAMESPACE_T(namespace_new());
}

void link_namespace_fill(struct link_namespace_t *ns)
{
  namespace_fill(&(ns->namespace));
}

void link_namespace_add_level(struct link_namespace_t *ns)
{
  namespace_add_level(&(ns->namespace), NS_LEVEL_T(link_level_new(ns)));
}

void link_namespace_del_level(struct link_namespace_t *ns)
{
  link_level_free(LINK_LEVEL_T(namespace_del_level(&(ns->namespace))));
}

int  link_namespace_is_empty(struct link_namespace_t *ns)
{
  return namespace_is_empty(&(ns->namespace));
}

void link_namespace_add(struct link_namespace_t *ns, char *name, struct expr_t *expr, int addr, int page, int instr_start)
{
  struct link_entry_t *entry;

  if (!ns->namespace.local)
    ERROR("internal error\n");
  entry = link_level_lookup_new(LINK_LEVEL_T(ns->namespace.local), name);
  if (link_entry_defined(entry))
    {
      WARNING("duplicate label: \"%s\", using first defined value\n", name);
      return;
    }
  link_entry_define(entry, LINK_LEVEL_T(ns->namespace.local), expr, addr, page, instr_start);
}

void link_namespace_lookup(struct link_namespace_t *ns, struct expr_t *expr, int addr, int page, int instr_start, int *dest)
{
  struct link_lookup_t *lookup;
  struct link_level_t *level;

  if (!(ns->namespace.local))
    ERROR("internal error\n");
  level = LINK_LEVEL_T(ns->namespace.local);

  lookup = link_lookup_new(level, expr, addr, page, instr_start, dest);
}

struct link_entry_t *link_namespace_lookup_direct(struct link_namespace_t *ns, char *name)
{
  struct link_level_t *level;
  struct link_entry_t **entry;

  if (!(ns->namespace.local))
    ERROR("internal error\n");
  for (level = LINK_LEVEL_T(ns->namespace.local); level; level = LINK_LEVEL_T(level->level.up))
    {
      entry = (struct link_entry_t **)hash_table_lookup(&(level->level.hash_table), name);
      if (*entry)
        {
	  if (link_entry_defined(*entry))
	    return *entry;
        }
    }
  return NULL;
}

void link_namespace_export(struct link_namespace_t *ns, char *name)
{
  struct link_entry_t *l_entry, *g_entry;
  struct link_global_t *global;
  
  struct link_level_t *level;

  level = LINK_LEVEL_T(ns->namespace.local);
  if (!level)
    ERROR("internal error\n");
  g_entry = link_level_lookup_new(LINK_LEVEL_T(ns->namespace.global), name);
  l_entry = link_level_lookup_new(level, name);

  if (link_entry_defined(g_entry) || link_entry_global(g_entry))
    {
      WARNING("duplicate label: \"%s\", using first defined value\n");
      return;
    }
  link_entry_set_global(g_entry);

  global = link_global_new(level, g_entry, l_entry);
  g_entry->expr = global->expr;

  if (link_entry_defined(l_entry))
    {
      link_namespace_do_export(ns, l_entry);
      return;
    }

  link_entry_set_global(g_entry);
}

void link_namespace_do_export(struct link_namespace_t *ns, struct link_entry_t *l_entry)
{
  struct link_entry_t *g_entry;

  if (!ns->namespace.global)
    ERROR("internal error\n");
  g_entry = link_level_lookup_new(LINK_LEVEL_T(ns->namespace.global), l_entry->entry.name);
  if (!link_entry_global(g_entry))
    ERROR("internal error\n");
  link_entry_define_global(g_entry, l_entry->page);
}

void link_namespace_delete(struct link_namespace_t *ns)
{
  namespace_delete(&(ns->namespace));
}

void link_namespace_free(struct link_namespace_t *ns)
{
  link_namespace_delete(ns);
  free(ns);
}

struct link_level_t *link_level_new(struct link_namespace_t *parent)
{
  struct link_level_t *val;

  ALLOCATE(struct link_level_t, val, 1);
  link_level_fill(val, parent);
  return val;
}

void link_level_fill(struct link_level_t *level, struct link_namespace_t *parent)
{
  ns_level_fill(&(level->level), &(parent->namespace));
  level->exprs = NULL;
  level->globals = NULL;
  level->expr_alloc = NULL;
  level->global_alloc = NULL;
  level->lookup_alloc = NULL;
  level->entry_alloc = NULL;
  level->string_alloc = NULL;
}

struct link_entry_t *link_level_lookup_new(struct link_level_t *level, char *name)
{
  struct link_entry_t **val;

  val = (struct link_entry_t **)hash_table_lookup(&(level->level.hash_table), name);
  if (*val == NULL)
    *val = link_entry_new(level, name);
  return *val;
}

void link_level_delete(struct link_level_t *level)
{
  link_global_copy(level->globals, LINK_LEVEL_T(level->level.up), LINK_LEVEL_T(level->level.parent->global));
  link_lookup_copy(level->exprs, LINK_LEVEL_T(level->level.up));

  link_expr_alloc_free(level->expr_alloc);
  link_global_alloc_free(level->global_alloc);
  link_lookup_alloc_free(level->lookup_alloc);
  link_entry_alloc_free(level->entry_alloc);
  string_alloc_free(level->string_alloc);
  ns_level_delete(&(level->level));
}

void link_level_free(struct link_level_t *level)
{
  link_level_delete(level);
  free(level);
}
