#define ZASM_FEEDER_C 1

/* feeder modification tokens
#token struct action_t *TOK_FEED_ACTION =
#token struct list_t *TOK_FEED_LIST =
*/

#include <zasm/feeder.h>
#include <zasm/memory.h>

struct feeder_t *feeder_new()
{
  struct feeder_t *val;
  ALLOCATE(struct feeder_t, val, 1);
  feeder_fill(val);
  return val;
}

void feeder_fill(struct feeder_t *feeder)
{
  feeder->source = NULL;
  feeder->backup = NULL;
}

void feeder_backup(struct feeder_t *feeder)
{
  struct feeder_src_t *src;

  src = feeder_src_new();
  src->type = FEEDER_SRC_LIST;
  src->previous = feeder->backup;
  src->content.list = list_new();
  feeder->backup = src;
}

void feeder_stop_backup(struct feeder_t *feeder)
{
  struct feeder_src_t *src;

  if (!feeder->backup)
    ERROR("internal error\n");
  src = feeder->backup;
  feeder->backup = src->previous;
  if (feeder->backup)
    list_add_list_end(feeder->backup->content.list, src->content.list);
  feeder_src_free(src);
}

void feeder_restore(struct feeder_t *feeder)
{
  struct feeder_src_t *src;

  if (!feeder->backup)
    ERROR("internal error\n");
  src = feeder->backup;
  feeder->backup = src->previous;
  src->previous = feeder->source;
  feeder->source = src;
}

void feeder_flush1(struct feeder_t *feeder, struct stream_t *output)
{
  struct feeder_src_t *src;

  if (!feeder->source)
    ERROR("internal error\n");
  list_add_list_end(LIST_T(output), feeder->source->content.list);
  src = feeder->source;
  feeder->source = src->previous;
  feeder_src_free(src);
}

void feeder_add_list(struct feeder_t *feeder, struct list_t *list)
{
  struct feeder_src_t *src;

  src = feeder_src_new();
  src->type = FEEDER_SRC_LIST;
  src->previous = feeder->source;
  src->content.list = list;
  feeder->source = src;
}

void feeder_add_action(struct feeder_t *feeder, struct action_t *action)
{
  struct feeder_src_t *src;
  struct stream_t *stream;

  src = feeder_src_new();
  src->type = FEEDER_SRC_ACTION;
  src->previous = feeder->source;
  stream = stream_new(action, NULL);
  action->output = stream;
  src->content.stream = stream;
  feeder->source = src;
}

#if 0
void feeder_add_feeder(struct feeder_t *feeder, struct feeder_t *other)
{
  struct feeder_src_t *src, *prev;
  
  if (!other->source)
    return;
  for (prev = NULL, src = other->source; src->previous; src = src->previous)
    prev = src;
  
  // stream is a descendant of list
  list_add_begin(feeder->source->content.list, src->content.list);
  feeder_src_free(src);
  if (prev)
    {
      prev->previous = feeder->source;
      feeder->source = other->source;
    }
  other->source = NULL;
  feeder_free(other);
}
#endif

struct token_t feeder_get(struct feeder_t *feeder)
{
  struct token_t token;
  struct feeder_src_t *src;

  if (!(feeder->source))
    ERROR("premature end of input\n");
  while (1)
    {
      token = feeder_src_get(feeder->source);
      if ((feeder->source) && (token.type == TOK_END_OF_FILE))
	{
	  src = feeder->source;
	  feeder->source = src->previous;
	  feeder_src_free(src);
	}
      else if (token.type == TOK_FEED_ACTION)
	feeder_add_action(feeder, token.value.t_struct_action_t_p);
      else if (token.type == TOK_FEED_LIST)
	feeder_add_list(feeder, token.value.t_struct_list_t_p);
      else
	break;
    }
  if (feeder->backup)
    list_add_end(feeder->backup->content.list, token);
  return token;
}

void feeder_unget(struct feeder_t *feeder, struct token_t token)
{
  if (token.type == TOK_END_OF_FILE)
    return;
  if (!feeder->source)
    feeder_add_list(feeder, list_new());
  feeder_src_unget(feeder->source, token);
  if (feeder->backup)
    list_get_end(feeder->backup->content.list);
}

void feeder_delete(struct feeder_t *feeder)
{
  if (feeder->source)
    ERROR("internal error\n");
  if (feeder->backup)
    ERROR("internal error\n");
}

void feeder_free(struct feeder_t *feeder)
{
  feeder_delete(feeder);
  free(feeder);
}

struct feeder_src_t *feeder_src_new()
{
  struct feeder_src_t *val;
  ALLOCATE(struct feeder_src_t, val, 1);
  feeder_src_fill(val);
  return val;
}

void feeder_src_fill(struct feeder_src_t *src)
{
  src->type = FEEDER_SRC_INVALID;
  src->previous = NULL;
}

struct token_t feeder_src_get(struct feeder_src_t *src)
{
  struct token_t token;

  if (!src)
    {
      token.type = TOK_END_OF_FILE;
      return token;
    }

  switch (src->type)
    {
    case FEEDER_SRC_LIST:
      token = list_get_begin(src->content.list);
      if (token.type == TOK_LIST_END)
	token.type = TOK_END_OF_FILE;
      return token;
    case FEEDER_SRC_ACTION:
      token = stream_get(src->content.stream);
      return token;
    default:
      ((struct feeder_src_t *)NULL)->type = 0;
      ERROR("internal error\n");
    }
  return token;
}

void feeder_src_unget(struct feeder_src_t *src, struct token_t token)
{
  switch (src->type)
    {
    case FEEDER_SRC_LIST:   return list_add_begin(src->content.list, token);
    case FEEDER_SRC_ACTION: return stream_unget(src->content.stream, token);
    default:
      ERROR("internal error\n");
    }
}

void feeder_src_delete(struct feeder_src_t *src)
{
  switch (src->type)
    {
    case FEEDER_SRC_LIST:
      list_free(src->content.list);
      break;
    case FEEDER_SRC_ACTION:
      action_free(src->content.stream->cb_read);
      stream_free(src->content.stream);
      break;
    default:
      ERROR("internal error\n");
    }
}

void feeder_src_free(struct feeder_src_t *src)
{
  if (!src)
    return;
  feeder_src_delete(src);
  free(src);
}
