#define ZASM_INTEL_HEX_WRITER_C 1

/*
#token void TOK_OUTPUT = /{DIR}output/ {} & {WRITE("#output ");}
*/

#include <zasm/intel_hex_writer.h>
#include <zasm/memory.h>
#include <stdio.h>
#include <stdarg.h>

struct intel_hex_writer_t *intel_hex_writer_new(char *filename, int endianness)
{
  struct intel_hex_writer_t *val;
  ALLOCATE(struct intel_hex_writer_t, val, 1);
  intel_hex_writer_fill(val, filename, endianness);
  return val;
}

void intel_hex_writer_fill(struct intel_hex_writer_t *writer, char *filename, int endianness)
{
  FILE *file;
  if (filename)
    {
      if (!(file = fopen(filename, "w")))
	ERROR("can not open \"%s\" for writing\n", filename);
    }
  else
    file = NULL;
  writer->file = file;
  writer->endianness = endianness;
  writer->length = 0;
  writer->type = 0;
  writer->address = 0;
  writer->page = -1;
  writer->checksum = 0;
}

static void intel_hex_writer_flush(struct intel_hex_writer_t *writer)
{
  int i, addr = 0;
  if (!(writer->file))
    {
      if (writer->length)
	ERROR("no output file specified\n");
      else
	return;
    }
  if ((writer->type == 0) && (writer->length == 0))
    return;
  if (writer->type == 0)
    {
      addr = writer->address;
      writer->address+=writer->length;
    }
  fprintf(writer->file, ":%02hhX%04hX%02hhX", writer->length, addr, writer->type);
  writer->checksum+=writer->length&0xff;
  writer->checksum+=addr&0xff;
  writer->checksum+=(addr>>8)&0xff;
  writer->checksum+=writer->type&0xff;
  for (i = 0; i < writer->length; i++)
    fprintf(writer->file, "%02hhX", writer->data[i]);
  fprintf(writer->file, "%02hhX\r\n", (0x100 - writer->checksum)&0xff);
  writer->checksum = 0;
  writer->length = 0;
  writer->type = 0;
}

static void intel_hex_writer_put_byte(struct intel_hex_writer_t *writer, unsigned char arg)
{
  if ((writer->length >= 32) || (writer->length < 0))
    ERROR("internal error\n");
  writer->data[writer->length] = arg;
  writer->checksum+=arg;
  writer->length++;
  if (writer->length >= 32)
    intel_hex_writer_flush(writer);
}

static int intel_hex_writer_check(int sign, char format, unsigned int value)
{
  switch (format)
    {
    case 'b':
    case 'B':
      if (sign)
	return token_value_signed_check(value,0xff);
      else
	return token_value_unsigned_check(value,0xff);
    case 'w':
    case 'W':
      if (sign)
	return token_value_signed_check(value,0xffff);
      else
	return token_value_unsigned_check(value,0xffff);
    case 'd':
    case 'D':
      if (sign)
	return token_value_signed_check(value,0xffffffff);
      else
	return token_value_unsigned_check(value,0xffffffff);
    default:
      return 0;
    }
}

// b: byte (8bit)
// w: word (16bit)
// d: double word (32bit)
// B: word (16bit, depend on endianness)
// W: word (16bit, depend on endianness)
// D: word (16bit, depend on endianness)
static void intel_hex_writer_writef(struct token_t t, struct intel_hex_writer_t *writer, char type, char *format, ...)
{
  va_list ap;
  unsigned int arg;
  int sign = 0;

  if (type != writer->type)
    intel_hex_writer_flush(writer);
  writer->type = type;
  va_start(ap, format);
  for (; *format; format++)
    {
      if (*format == 's')
	{
	  sign=1;
	  continue;
	}
      arg = va_arg(ap, unsigned int);
      if (!intel_hex_writer_check(sign, *format, arg))
	T_WARNING(t, "value out of bounds\n");
      switch (*format)
	{
	case 'd':
	  intel_hex_writer_put_byte(writer,(arg>>24)&0xff);
	  intel_hex_writer_put_byte(writer,(arg>>16)&0xff);
	case 'w':
	  intel_hex_writer_put_byte(writer,(arg>>8)&0xff);
	case 'b':
	  intel_hex_writer_put_byte(writer,arg&0xff);
	  break;
	case 'D':
	  if (writer->endianness == PROC_LITTLE_ENDIAN)
	    {
	      intel_hex_writer_put_byte(writer, arg&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>8)&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>16)&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>24)&0xff);
	    }
	  else if (writer->endianness == PROC_BIG_ENDIAN)
	    {
	      intel_hex_writer_put_byte(writer, (arg>>24)&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>16)&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>8)&0xff);
	      intel_hex_writer_put_byte(writer, arg&0xff);
	    }
	  break;
	case 'W':
	  if (writer->endianness == PROC_LITTLE_ENDIAN)
	    {
	      intel_hex_writer_put_byte(writer, arg&0xff);
	      intel_hex_writer_put_byte(writer, (arg>>8)&0xff);
	    }
	  else if (writer->endianness == PROC_BIG_ENDIAN)
	    {
	      intel_hex_writer_put_byte(writer, (arg>>8)&0xff);
	      intel_hex_writer_put_byte(writer, arg&0xff);
	    }
	  break;
	case 'B':
	  intel_hex_writer_put_byte(writer,arg&0xff);
	  break;
	}
      sign=0;
    }
  va_end(ap);
}

static void intel_hex_writer_close(struct token_t t, struct intel_hex_writer_t *writer)
{
  if (writer->file)
    {
      intel_hex_writer_writef(t, writer, 1, "");
      intel_hex_writer_flush(writer);
      if (fclose(writer->file) < 0)
	ERROR("I/O error\n");
      writer->file = NULL;
    }
}

void intel_hex_writer_execute(struct feeder_t *input, struct stream_t *output, struct intel_hex_writer_t *writer)
{
  struct token_t token;
  FILE *file;

  for (token = feeder_get(input); token.type != TOK_END_OF_FILE; token = feeder_get(input))
    {
      switch (token.type)
	{
	case TOK_OUTPUT:
	  token = feeder_get(input);
	  while (token.type == TOK_SPACE)
	    token = feeder_get(input);
	  if (token.type != TOK_STRING)
	    T_ERROR(token, "syntax error\n");
	  if (!(file = fopen(token.value.t_char_p, "w")))
	    ERROR("can not open \"%s\" for writing\n", token.value.t_char_p);
	  token_free(token);
	  intel_hex_writer_close(token, writer);
	  writer->file = file;
	  break;
	case TOK_SET_PAGE:
	  intel_hex_writer_flush(writer);
	  if (token.value.t_struct_page_addr_t.page != writer->page)
	    {
	      writer->page = token.value.t_struct_page_addr_t.page;
	      intel_hex_writer_writef(token, writer, 2, "w", writer->page);
	    }
	  writer->address = token.value.t_struct_page_addr_t.addr;
	  break;
	case TOK_BYTE_INT:
	  intel_hex_writer_writef(token, writer, 0, "B", token.value.t_int);
	  break;
	case TOK_BYTE_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "B", (*(token.value.t_int_p)));
	  break;
	case TOK_WORD_INT:
	  intel_hex_writer_writef(token, writer, 0, "W", token.value.t_int);
	  break;
	case TOK_WORD_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "W", (*(token.value.t_int_p)));
	  break;
	case TOK_DWORD_INT:
	  intel_hex_writer_writef(token, writer, 0, "D", token.value.t_int);
	  break;
	case TOK_DWORD_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "D", (*(token.value.t_int_p)));
	  break;
	case TOK_SIGNED_BYTE_INT:
	  intel_hex_writer_writef(token, writer, 0, "sB", token.value.t_int);
	  break;
	case TOK_SIGNED_BYTE_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "sB", (*(token.value.t_int_p)));
	  break;
	case TOK_SIGNED_WORD_INT:
	  intel_hex_writer_writef(token, writer, 0, "sW", token.value.t_int);
	  break;
	case TOK_SIGNED_WORD_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "sW", (*(token.value.t_int_p)));
	  break;
	case TOK_SIGNED_DWORD_INT:
	  intel_hex_writer_writef(token, writer, 0, "sD", token.value.t_int);
	  break;
	case TOK_SIGNED_DWORD_INT_P:
	  intel_hex_writer_writef(token, writer, 0, "sD", (*(token.value.t_int_p)));
	  break;
	default:
	  break;
	}
    }
  intel_hex_writer_close(token, writer);
  feeder_free(input);
  intel_hex_writer_free(writer);
}

void intel_hex_writer_delete(struct intel_hex_writer_t *writer)
{
  if (writer->file)
    if (fclose(writer->file) < 0)
      ERROR("I/O error\n");
}

void intel_hex_writer_free(struct intel_hex_writer_t *writer)
{
  intel_hex_writer_delete(writer);
  free(writer);
}
