#define ZASM_WRITER_C 1

#include <zasm/writer.h>
#include <zasm/memory.h>

#define WRITE(fmt, args...) fprintf(writer->output, fmt, ## args)
#define CHECK_SIGNED(VAL,MASK) (({if (!token_value_signed_check(VAL,MASK)) T_WARNING(token,"value out of bounds\n");}))
#define CHECK_UNSIGNED(VAL,MASK) (({if (!token_value_unsigned_check(VAL,MASK)) T_WARNING(token,"value out of bounds\n");}))

struct writer_t *writer_new(char *filename)
{
  struct writer_t *val;
  ALLOCATE(struct writer_t, val, 1);
  writer_fill(val, filename);
  return val;
}

void writer_fill(struct writer_t *writer, char *filename)
{
  FILE *file;
  file = fopen(filename, "w");
  if (!file)
    ERROR("can not open file \"%s\"\n", filename);
  writer->output = file;
}

void writer_change(struct writer_t *writer, char *filename)
{
  writer_delete(writer);
  writer_fill(writer, filename);
}

void writer_execute(struct feeder_t *input, struct stream_t *output, struct writer_t *writer)
{
  struct token_t token;

  for (token = feeder_get(input); token.type != TOK_END_OF_FILE; token = feeder_get(input))
    {
      switch (token.type)
	{
case TOK_MULT_LIST: break;
case TOK_PROCESSOR: break;
case TOK_PROC_INSTR: break;
case TOK_PROC_INSTR_ARG: break;
case TOK_LITTLE_ENDIAN: {WRITE("LITTLE ENDIAN");}break;
case TOK_BIG_ENDIAN: {WRITE("BIG ENDIAN");}break;
case TOK_ADDINSTR: {WRITE("#addinstr ");}break;
case TOK_BYTE_INT: {CHECK_UNSIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_WORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_DWORD_INT: {CHECK_UNSIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_BYTE_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_WORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_DWORD_INT_P: {CHECK_UNSIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_BYTE_INT: {CHECK_SIGNED(token.value.t_int,0xff); WRITE("`%02hhX`", token.value.t_int);}break;
case TOK_SIGNED_WORD_INT: {CHECK_SIGNED(token.value.t_int,0xffff); WRITE("`%04hX`", token.value.t_int);}break;
case TOK_SIGNED_DWORD_INT: {CHECK_SIGNED(token.value.t_int,0xffffffff); WRITE("`%08lX`", token.value.t_int);}break;
case TOK_SIGNED_BYTE_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xff); WRITE("`%02hhX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_WORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffff); WRITE("`%04hX`", *(token.value.t_int_p));}break;
case TOK_SIGNED_DWORD_INT_P: {CHECK_SIGNED(*(token.value.t_int_p),0xffffffff); WRITE("`%08lX`", *(token.value.t_int_p));}break;
case TOK_BYTE: {WRITE("#byte ");}break;
case TOK_WORD: {WRITE("#word ");}break;
case TOK_DWORD: {WRITE("#dword ");}break;
case TOK_SIGNED_BYTE: {WRITE("#sbyte ");}break;
case TOK_SIGNED_WORD: {WRITE("#sword ");}break;
case TOK_SIGNED_DWORD: {WRITE("#sdword ");}break;
case TOK_ARG: {WRITE("@%i", token.value.t_int);}break;
case TOK_DB: {WRITE(".db ");}break;
case TOK_DW: {WRITE(".dw ");}break;
case TOK_DD: {WRITE(".dd ");}break;
case TOK_NONE: break;
case TOK_LIST_END: break;
case TOK_STREAM_END: break;
case TOK_END_OF_FILE: break;
case TOK_STRING: {WRITE("\"%s\"", token.value.t_char_p);}break;
case TOK_ISTRING: {WRITE("<%s>", token.value.t_char_p);}break;
case TOK_NAME: {WRITE("%s", token.value.t_char_p);}break;
case TOK_VARARG: {WRITE("%s...", token.value.t_char_p);}break;
case TOK_NEWLINE: {WRITE("\n");}break;
case TOK_ESCAPED_NEWLINE: {WRITE("\n");}break;
case TOK_SPACE: {WRITE(" ");}break;
case TOK_SPECIAL: {WRITE("%c", token.value.t_char);}break;
case TOK_INTEGER: {WRITE("%i", token.value.t_int);}break;
case TOK_BEGIN_INSTR: {WRITE("#bi ");}break;
case TOK_ALIGN_ADDR: {WRITE("#align_addr ");}break;
case TOK_ALIGN_MULT: {WRITE("#align_mult ");}break;
case TOK_ORG: {WRITE("#org ");}break;
case TOK_PAGE: {WRITE("#setpage ");}break;
case TOK_PAGE_START: {WRITE("#page_start ");}break;
case TOK_NEWPAGE: {WRITE("#newpage ");}break;
case TOK_EQU: {WRITE("#equ ");}break;
case TOK_EXPORT: {WRITE("#export ");}break;
case TOK_BEGIN_SECTION: {WRITE("#begin_section ");}break;
case TOK_END_SECTION: {WRITE("#end_section ");}break;
case TOK_SHARP: {WRITE("#");}break;
case TOK_BLOCK: {WRITE("#block ");}break;
case TOK_FILL: {WRITE("#fill ");}break;
case TOK_SET_PAGE: break;
case TOK_PC_FAIL: {WRITE("<pc fail>");}break;
case TOK_PC_FAIL1: {WRITE("<pc fail1>");}break;
case TOK_PC_SUCCESS: {WRITE("<pc success>");}break;
case TOK_BACKSLASH: {WRITE("\\");}break;
case TOK_EXPR: {WRITE("<expr>");}break;
case TOK_LIST: {WRITE("<list>");}break;
case TOK_END_FILE: {WRITE("<endFile>");}break;
case TOK_OP_IFTHEN: {WRITE("?");}break;
case TOK_OP_ELSE: {WRITE(":");}break;
case TOK_OP_LOG_OR: {WRITE("||");}break;
case TOK_OP_LOG_AND: {WRITE("&&");}break;
case TOK_OP_LOG_XOR: {WRITE("^^");}break;
case TOK_OP_LOG_NOT: {WRITE("!");}break;
case TOK_OP_BIN_OR: {WRITE(".|");}break;
case TOK_OP_BIN_AND: {WRITE(".&");}break;
case TOK_OP_BIN_XOR: {WRITE(".^");}break;
case TOK_OP_BIN_NOT: {WRITE(".~");}break;
case TOK_OP_SHIFT_LEFT: {WRITE("<<");}break;
case TOK_OP_SHIFT_RIGHT: {WRITE(">>");}break;
case TOK_OP_EQU: {WRITE("==");}break;
case TOK_OP_NEQ: {WRITE("!=");}break;
case TOK_OP_LT: {WRITE("<");}break;
case TOK_OP_GT: {WRITE(">");}break;
case TOK_OP_LE: {WRITE("<=");}break;
case TOK_OP_GE: {WRITE(">=");}break;
case TOK_OP_CMP: {WRITE("<=>");}break;
case TOK_OP_ADD: {WRITE("+");}break;
case TOK_OP_SUB: {WRITE("-");}break;
case TOK_OP_MUL: {WRITE("*");}break;
case TOK_OP_DIV: {WRITE("/");}break;
case TOK_OP_MOD: {WRITE("%%");}break;
case TOK_OP_POW: {WRITE("**");}break;
case TOK_DOT: {WRITE(".");}break;
case TOK_PAREN_OPEN: {WRITE("(");}break;
case TOK_PAREN_CLOSE: {WRITE(")");}break;
case TOK_BRACE_OPEN: {WRITE("{");}break;
case TOK_BRACE_CLOSE: {WRITE("}");}break;
case TOK_SQUARE_OPEN: {WRITE("[");}break;
case TOK_SQUARE_CLOSE: {WRITE("]");}break;
case TOK_COMMA: {WRITE(", ");}break;
case TOK_PASTE: {WRITE(" ## ");}break;
case TOK_DEFINED: {WRITE("defined");}break;
case TOK_DEFINE: {WRITE("#define");}break;
case TOK_MACRO: {WRITE(".macro");}break;
case TOK_REDEF: {WRITE("#redef");}break;
case TOK_UNDEF: {WRITE("#undef");}break;
case TOK_ENDM: {WRITE("#endm");}break;
case TOK_PP_BEGIN_SECTION: {WRITE("#pp_begin_section");}break;
case TOK_PP_END_SECTION: {WRITE("#pp_end_section");}break;
case TOK_INCLUDE: {WRITE("#include");}break;
case TOK_IF: {WRITE("#if");}break;
case TOK_THEN: {WRITE("#then");}break;
case TOK_IFDEF: {WRITE("#ifdef");}break;
case TOK_IFNDEF: {WRITE("#ifndef");}break;
case TOK_ELSE: {WRITE("#else");}break;
case TOK_ELSEIF: {WRITE("#elseif");}break;
case TOK_ELSEIFDEF: {WRITE("#elseifdef");}break;
case TOK_ELSEIFNDEF: {WRITE("#elseifndef");}break;
case TOK_ENDIF: {WRITE("#endif");}break;
case TOK_FOR: {WRITE("#for");}break;
case TOK_IN: {WRITE("#in");}break;
case TOK_FROM: {WRITE("#from");}break;
case TOK_TO: {WRITE("#to");}break;
case TOK_STEP: {WRITE("#step");}break;
case TOK_DO: {WRITE("#do");}break;
case TOK_ENDFOR: {WRITE("#endfor");}break;
case TOK_ERROR: {WRITE("#error");}break;
case TOK_WARNING: {WRITE("#warning");}break;
case TOK_DEBUG: {WRITE("#debug");}break;
case TOK_FEED_ACTION: break;
case TOK_FEED_LIST: break;
case TOK_OUTPUT: {WRITE("#output ");}break;

	default:
	  ERROR("unknown token\n"); break;
	}
    }
  writer_free(writer);
  return;
}

void writer_delete(struct writer_t *writer)
{
  if (fclose(writer->output))
    ERROR("I/O error\n");
}

void writer_free(struct writer_t *writer)
{
  writer_delete(writer);
  free(writer);
}
